--- src/Exmap.cpp.orig	Thu Sep 28 17:52:25 2006
+++ src/Exmap.cpp	Mon Jan 14 15:37:56 2008
@@ -1,17 +1,39 @@
 /*
  * (c) John Berthels 2005 <jjberthels@gmail.com>. See COPYING for license.
+ *
+ * vim:set sts=4 ts=8 sw=4:
  */
-#include "jutil.hpp"
-#include "Exmap.hpp"
-#include "Elf.hpp"
-
 #include <sstream>
 #include <set>
 
-#include <ctype.h>
 #include <sys/types.h>
+
+#ifdef __FreeBSD__
+#include <sys/param.h>
+#include <sys/sysctl.h>
+#include <sys/syscall.h>
+#include <sys/module.h>
+#include <sys/user.h>
+#include <sys/vnode.h>
+#include <sys/linker.h>
+
+#include <vm/vm_object.h>
+#include <vm/swap_pager.h>
+#endif
+
+#include <ctype.h>
 #include <unistd.h>
 
+#ifdef __FreeBSD__
+#include <fcntl.h>
+#include <kvm.h>
+#include "pmap_helper.h"
+#endif
+
+#include "jutil.hpp"
+#include "Exmap.hpp"
+#include "Elf.hpp"
+
 using namespace Exmap;
 using namespace std;
 using namespace jutil;
@@ -959,6 +981,324 @@
 SysInfo::~SysInfo()
 { }
 
+#ifdef __FreeBSD__
+
+// The FreeBSD code is believed to return accurate page counts,
+// however the resident page information may not truly reflect
+// the layout of the mappings in physical memory. This is due to
+// API limitations.
+
+FbsdSysInfo::FbsdSysInfo()
+ : _page_cookie(1),
+   _modid(-1),
+   _pmap_helper_syscall_num(-1),
+   _kd(0)
+{
+    // Check that kernel module support is loaded; if not, try to load it.
+    int tmp_modid = modfind("pmap_helper");
+    if (tmp_modid == -1) {
+	warn << "pmap_helper module not found, attempting to load it.\n";
+	tmp_modid = kldload("pmap_helper.ko");
+	if (tmp_modid == -1)
+	    warn << "unable to load pmap_helper module.\n";
+	tmp_modid = modfind("pmap_helper");
+	if (tmp_modid == -1)
+	    warn << "pmap_helper module loaded but not found.\n";
+    }
+
+    _modid = tmp_modid;
+
+    if (tmp_modid != -1) {
+	struct module_stat stat;
+
+	stat.version = sizeof(stat);
+	modstat(tmp_modid, &stat);
+	_pmap_helper_syscall_num = stat.data.intval;
+    }
+
+    // Try to open kvm.
+    kvm_t* tmp_kd = kvm_openfiles(NULL, NULL, NULL, O_RDONLY, _kvm_errbuf);
+    if (tmp_kd == 0) {
+	warn << "kvm_openfiles: " << _kvm_errbuf << "\n";
+    }
+    _kd = tmp_kd;
+}
+
+FbsdSysInfo::~FbsdSysInfo()
+{
+    if (_kd != 0) {
+	int retval = kvm_close(_kd);
+	if (retval == -1)
+	    warn << "kvm_close: " << kvm_geterr(_kd) << "\n";
+	_kd = 0;
+    }
+}
+
+list<pid_t> FbsdSysInfo::accessible_pids()
+{
+    list<pid_t> pids;
+    int nprocs;
+
+    // get all running pids in system.
+
+    if (_kd == 0) {
+	warn << "_kd is NULL\n";
+	return pids;
+    }
+
+    struct kinfo_proc *kps = kvm_getprocs(_kd, KERN_PROC_PROC, 0, &nprocs);
+    if (kps == NULL) {
+	warn << "kvm_getprocs: " << kvm_geterr(_kd) << "\n";
+	return pids;
+    }
+
+    struct kinfo_proc *kp = kps;
+    for (int i = 0; i < nprocs; i++, kp++) {
+	pids.push_back(kp->ki_pid);
+    }
+
+    return pids;
+}
+
+bool FbsdSysInfo::sanity_check()
+{
+    return _modid != -1 &&
+	   _pmap_helper_syscall_num != -1 &&
+	   _kd != 0;
+}
+
+bool FbsdSysInfo::read_page_info(pid_t pid,
+				  map<Address, list<Page> > &page_info)
+{
+    page_info.clear();
+
+    if (_kd == 0) {
+	warn << "_kd is NULL\n";
+	return false;
+    }
+
+    struct pmh_map *maps = (struct pmh_map *)
+	calloc(MAX_MAPS, sizeof(struct pmh_map));
+    if (maps == NULL) {
+	warn << "calloc: " << strerror(errno) << "\n";
+	return false;
+    }
+
+    struct pmh pmh;
+    pmh.version = PMH_VERSION;
+    pmh.pid = pid;
+    pmh.nmaps = MAX_MAPS;
+    pmh.maps = maps;
+    if (syscall(_pmap_helper_syscall_num, &pmh) != 0) {
+	warn << "pmap_helper_syscall: " << strerror(errno) << "\n";
+	free(maps);
+	return false;
+    }
+
+    list<Page> empty_pagelist;
+    list<Page> *current_pagelist = NULL;
+
+    size_t nmaps = pmh.nmaps;
+    size_t page_size = Elf::page_size();
+
+    // for each vm_entry in proc vmspace:
+    //  fill out the map of Address to page list using a second pass
+    //  over pmh_maps.
+    struct pmh_map *pm = maps;
+    for (size_t i = 0; i < nmaps; i++, pm++) {
+	if (!pm->lobj_present)
+	    continue;
+
+	// Build a page list for each map entry.
+	size_t npages = pm->entry.end - pm->entry.start / page_size;
+	page_info[pm->entry.start] = empty_pagelist;
+	current_pagelist = &(page_info[pm->entry.start]);
+
+	size_t nresident = 0;
+
+	// Determine resident page counts for this map entry;
+	// dependent upon which pager the FreeBSD VM used for this
+	// entry's lowest stacked VM object.
+	switch (pm->lobj.type) {
+	case OBJT_DEFAULT:
+	    nresident = pm->resident_pages;
+	    break;
+	case OBJT_VNODE:
+	    if (pm->obj.shadow_count == 1) {
+		// This vm_map_entry represents a process-private mapping;
+		// there's a shadow object.
+		nresident = pm->obj.resident_page_count;
+	    } else {
+		nresident = pm->resident_pages;
+	    }
+	    break;
+	case OBJT_DEVICE:   // not counted as resident for Exmap purposes.
+	case OBJT_SWAP:	    // underlying object is swap (paged out).
+	case OBJT_DEAD:	    // dead.
+	default:	    // unknown.
+	    nresident = 0;
+	    break;
+	}
+
+	size_t rresident = 0;
+
+	// For each page in entry mapping, fake up a unique page descriptor
+	// internal to Exmap which records if the pages are writable,
+	// and fake up the appropriate resident pages (we do this at
+	// the start of the mapping).
+	//
+	// We can't tell for sure if the page is resident
+	// or writable, or even know its address, without going to pmap;
+	// see comments elsewhere about this limitation.
+	//
+	for (size_t j = 0; j < npages; j++) {
+	    Page p(get_page_cookie(),
+		   rresident++ <= nresident,
+		   (pm->entry.protection & VM_PROT_WRITE));
+	    current_pagelist->push_back(p);
+	}
+    }
+
+    free(maps);
+    return true;
+}
+
+std::string FbsdSysInfo::read_cmdline(pid_t pid)
+{
+    string cmdline;
+    int nprocs;
+
+    if (_kd == 0) {
+	warn << "_kd is NULL\n";
+	return cmdline;
+    }
+
+    // get the command line for a specific pid.
+
+    struct kinfo_proc *kps = kvm_getprocs(_kd, KERN_PROC_PID, pid, &nprocs);
+    if (kps == NULL) {
+	warn << "kvm_getprocs: " << kvm_geterr(_kd) << "\n";
+	return cmdline;
+    }
+
+    if (nprocs > 1) {
+	warn << "kvm_getprocs() returned " << nprocs << "\n";
+    }
+
+    if (kps->ki_args != NULL) {
+	struct pargs *pa = NULL;
+
+	pa = (struct pargs *)realloc(pa, sizeof(struct pargs));
+	ssize_t ret = kvm_read(_kd, (u_long)kps->ki_args, pa,
+			       sizeof(struct pargs));
+	if (ret != sizeof(struct pargs)) {
+	    warn << "kvm_read(): " << kvm_geterr(_kd) << "\n";
+	} else {
+	    ssize_t new_size = sizeof(struct pargs) + pa->ar_length;
+	    pa = (struct pargs *)realloc(pa, new_size);
+	    ret = kvm_read(_kd, (u_long)kps->ki_args, pa, new_size);
+	    cmdline = (char *)pa->ar_args;
+	}
+	if (pa != NULL)
+	    free(pa);
+    } else {
+	cmdline = kps->ki_comm;
+    }
+
+    return cmdline;
+}
+
+const int FbsdSysInfo::MAX_MAPS = 1024;
+
+bool FbsdSysInfo::read_vmas(const PagePoolPtr &pp,
+			     pid_t pid,
+			     list<VmaPtr> &vmas)
+{
+    vmas.clear();
+
+    if (_kd == 0) {
+	warn << "_kd is NULL\n";
+	return false;
+    }
+
+    // get VMA entries for pid.
+
+    struct pmh_map *maps = (struct pmh_map *)
+	calloc(MAX_MAPS, sizeof(struct pmh_map));
+    if (maps == NULL) {
+	warn << "calloc: " << strerror(errno) << "\n";
+	return false;
+    }
+
+    struct pmh pmh;
+    pmh.version = PMH_VERSION;
+    pmh.pid = pid;
+    pmh.nmaps = MAX_MAPS;
+    pmh.maps = maps;
+    if (syscall(_pmap_helper_syscall_num, &pmh) != 0) {
+	warn << "pmap_helper_syscall: " << strerror(errno) << "\n";
+	free(maps);
+	return false;
+    }
+
+    string fname;
+    size_t nmaps = pmh.nmaps;
+
+    // for each vm_entry in proc vmspace:
+    //  determine size, location, type and name of mapping.
+    //  page counts are calculated in the read_page_info() pass.
+    struct pmh_map *pm = maps;
+    for (size_t i = 0; i < nmaps; i++, pm++) {
+	if (!pm->lobj_present)
+	    continue;
+
+	switch (pm->lobj.type) {
+	case OBJT_DEFAULT:
+	    fname = "[anon]";
+	    break;
+	case OBJT_VNODE:
+	    fname = (char *)pm->fname;
+	    break;
+	case OBJT_SWAP:
+	    fname = "[swap]";
+	    break;
+	case OBJT_DEVICE:
+	    fname = "/dev/";
+	    fname += (char *)pm->fname;
+	    break;
+	case OBJT_DEAD:
+	    fname = "[dead]";
+	    break;
+	default:
+	    fname = "[unknown]";
+	    break;
+	}
+
+	VmaPtr vma;
+	vma.reset(new Vma(pm->entry.start, pm->entry.end, pm->entry.offset,
+			  fname));
+	vma->selfptr(vma);
+	vmas.push_back(vma);
+    }
+
+    free(maps);
+    return true;
+}
+
+//
+// XXX: We can't retrieve the physical address of an object's
+// pages without delving into pmap, and there is no guarantee
+// it will remain the same during a snapshot unless we take
+// the pmap lock in pmap_helper. So to avoid modifying pmap_helper,
+// we just fake it up here.
+//
+PageCookie FbsdSysInfo::get_page_cookie()
+{
+    return ++_page_cookie;
+}
+
+#else // !defined(__FreeBSD__)
+
 LinuxSysInfo::~LinuxSysInfo()
 { }
 
@@ -1181,7 +1521,7 @@
     return sstr.str();
 }
 
-
+#endif // defined(__FreeBSD__)
 
     
 MapCalculator::MapCalculator(const list<VmaPtr> &vmas,
